#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <termios.h>
#include <fcntl.h>
#include <string.h>
#include <sys/shm.h>
#include <sys/ipc.h>
#include <dtk.h>
#include <dtk/dtkService.h>

// this works on my Dell laptop
//#define DEVICE_FILE   "/dev/ttyS0"

// this works on visage, the NIST Onyx
#define DEVICE_FILE   "/dev/ttyd4"

/*
#define TRACKD_COMPATABLE
*/

#define READ_BUFFER_SIZE        ((size_t) 128)

// clamp joystick numbers to this absolute value
#define MAX_XY_OFFSET (32.f)

class Vwand : public dtkService {
public:
  Vwand(void) ;
  virtual ~Vwand(void) ;
  int serve() ;
  
private:
  
  void clean_up(void) ;
  
  dtkSharedMem *button_shm ;
  dtkSharedMem *joystick_shm ;
  
  float joystick_state[2] ;
  unsigned char button_state ;
#ifdef TRACKD_COMPATABLE 
  dtkVRCOSharedMem VRCO ;
#endif
  int termio_set_flag ;
  struct termios original_terminfo ;
  
  unsigned char read_buffer[READ_BUFFER_SIZE] ;

} ;


void Vwand::clean_up(void) {
  if(button_shm) {
    delete button_shm ;
    button_shm = NULL ;
  }
  
  if(joystick_shm) {
    delete joystick_shm ;
    joystick_shm = NULL ;
  }
  
  if(fd > -1) {
    if(termio_set_flag) {
      termio_set_flag = 0 ;
      if (tcsetattr(fd, TCSANOW, &original_terminfo) == -1)
	dtkMsg.add(DTKMSG_WARN, 1,
		   "Vwand::~Vwand() failed: failed "
		   "to reset serial device"
		   " termio settings: "
		   "tcsetattr(%d,,) failed.\n",fd) ;
    }
    close(fd) ;
    fd = -1 ;
  }
}


Vwand::~Vwand(void) {
  clean_up() ;
}


Vwand::Vwand(void) {
  button_shm = NULL ;
  joystick_shm = NULL ;
  joystick_state[0] = 0.0f ;
  joystick_state[1] = 0.0f ;
  termio_set_flag = 0 ;
  
#ifdef TRACKD_COMPATABLE
  if(VRCO.isInvalid()) return ; // error
#endif
  
  /************************************************************/
  /***************** open serial device file ******************/
  /************************************************************/
  
  if((fd = open(DEVICE_FILE, O_RDWR|O_NONBLOCK)) < 0) {
    dtkMsg.add(DTKMSG_ERROR, 1,
	       "Vwand::Vwand() failed: open(\"%s\", "
	       "O_RDWR|O_NONBLOCK)"
	       " failed.\n", DEVICE_FILE) ;
    clean_up() ;
    return ;
  }
  
  // get and save original termio info
  if(tcgetattr(fd, &original_terminfo) == -1) {
    dtkMsg.add(DTKMSG_ERROR, 1,
	       "Vwand::Vwand() failed: "
	       "tcgetattr(%d,,) failed.\n",
	       fd) ;
    clean_up() ;
    return ;
  }
  
  struct termios terminfo ;
  terminfo.c_iflag = 0 ;
  terminfo.c_oflag = 0 ;
  terminfo.c_lflag = 0 ;
  
  terminfo.c_cflag = CS7 | HUPCL | CSTOPB | CREAD | CLOCAL ;
  if (cfsetospeed(&terminfo, B1200)  == -1) {
    dtkMsg.add(DTKMSG_ERROR, 1,
	       "Vwand::Vwand() failed: "
	       "cfsetospeed(%d,,) failed.\n",fd) ;
    clean_up() ;
    return ;
  }
  
  if (cfsetispeed(&terminfo, B1200)  == -1) {
    dtkMsg.add(DTKMSG_ERROR, 1,
	       "Vwand::Vwand() failed: "
	       "cfsetispeed(%d,,) failed.\n",fd) ;
    clean_up() ;
    return ;
  }
  
  for (int i=0 ; i<NCCS ; i++)
    terminfo.c_cc[i] = 0 ;
  
  if (tcsetattr(fd, TCSANOW, &terminfo) == -1) {
    dtkMsg.add(DTKMSG_ERROR, 1,
	       "Vwand::Vwand() failed: "
	       "tcsetattr(%d,,) failed.\n",fd) ;
    clean_up() ;
    return ;
  }
  termio_set_flag = 1 ;
  
  if (tcflush(fd, TCIOFLUSH) == -1) {
    dtkMsg.add(DTKMSG_ERROR, 1,
	       "Vwand::Vwand() failed: "
	       "Could not flush serial port: "
	       "tcflush() failed.\n") ;
    clean_up() ;
    return ;
  }
  
  /*************************************************************/
  /****************** get some DTK shared memory ***************/
  /*************************************************************/
  
  button_shm = new
    dtkSharedMem(sizeof(unsigned char),"buttons", &button_state) ;
  joystick_shm = new
    dtkSharedMem(sizeof(float)*2,"joystick", joystick_state) ;
  
  if(!button_shm || !joystick_shm) {
    // This will let you know why dtkSharedMem::dtkSharedMem()
    // spewed.
    dtkMsg.add(DTKMSG_ERROR,
	       "Vwand::Vwand() failed.\n") ;
    clean_up() ;
    return ;
  }
}


int Vwand::serve(void) {
  ssize_t bytes_read =
    read(fd, read_buffer, READ_BUFFER_SIZE) ;
  
  /*************************************************************/
  /*** checking error cases first. Cause it's easier that way **/
  /*************************************************************/
  
  if(bytes_read <= 0) {
    if(bytes_read == 0) { // This should never happen.
      dtkMsg.add(DTKMSG_ERROR, 1,
                 "Vwand::serve() failed: "
                 "read() read no data.\n") ;
      return DTKSERVICE_CONTINUE ;
    } else { // if(bytes_read < 0) // error
      dtkMsg.add(DTKMSG_ERROR, 1,
		 "Vwand::serve() failed: "
		 "read() failed.\n") ;
      return DTKSERVICE_CONTINUE ;
    }
  }
  
  /*************************************************************/
  /***************** process the data read *********************/
  /*************************************************************/
  
  static bool gotStatusByte1 = false ;
  static bool gotStatusByte2 = false ;
  static int byteNum = 1 ;
  static bool left = 0 ;
  static bool right = 0 ;
  static bool middle = 0 ;
  bool bit ;
  static signed char dx, dy ;

  for(unsigned char *ptr=read_buffer ;ptr < read_buffer+bytes_read ;ptr++) {
    if (!gotStatusByte1 && !gotStatusByte2 && *ptr==0x4D) {
      gotStatusByte1 = true ;
      //printf("gotStatusByte1\n") ;

    } else if (gotStatusByte1 && !gotStatusByte2 && *ptr==0x33) {
      gotStatusByte2 = true ;
      //printf("gotStatusByte2\n") ;

    } else if (gotStatusByte1 && gotStatusByte2) {

      if ((*ptr & 0x40) && ((byteNum == 1) || (byteNum == 4))) {
	//printf("    dx = %d\n", dx) ;
	//printf("    dy = %d\n", dy) ;
	//printf("joystick = (%+3.3f %+3.3f) buttons = 0x%2x\n",
	//joystick_state[0], joystick_state[1], button_state) ;
	//printf("  Byte 1 = 0x%2.2x\n",*ptr) ;
	if (left != (bit = (*ptr & 0x10))) {
	  left = bit ;
	  button_state &= ~(0x01) ;
	  button_state |= (*ptr & 0x10)>>4 ;
	  button_shm->write(&button_state);
	  //printf("left = %d\n\n\n",left) ;
	}
	if (right != (bit = (*ptr & 0x20))) {
	  right = bit ;
	  button_state &= ~(0x04) ;
	  button_state |= (*ptr & 0x20)>>3 ;
	  button_shm->write(&button_state);
	  //printf("right = %d\n\n",right) ;
	}
	dx = dy = 0 ;
	dx |= (*ptr & 0x3)<<6 ;
	dy |= (*ptr & 0xC)<<4 ;
	byteNum = 2 ;

      } else if (!(*ptr & 0x40) && (byteNum == 2)) {
	//printf("  Byte 2 = 0x%2.2x\n",*ptr) ;
	dx |= (*ptr & 0x3F) ;
	joystick_state[0] = -(float)dx/MAX_XY_OFFSET ;
	if (joystick_state[0]>1.f)
	  joystick_state[0] = 1.f ;
	else if (joystick_state[0]<-1.f)
	  joystick_state[0] = -1.f ;
	byteNum = 3 ;

      } else if (!(*ptr & 0x40) && (byteNum == 3)) {
	//printf("  Byte 3 = 0x%2.2x\n",*ptr) ;
	dy |= (*ptr & 0x3F) ;
	joystick_state[1] = (float)dy/MAX_XY_OFFSET ;
	if (joystick_state[1]>1.f)
	  joystick_state[1] = 1.f ;
	else if (joystick_state[1]<-1.f)
	  joystick_state[1] = -1.f ;
	joystick_shm->write(joystick_state);
	//printf("   dx, dy = %+3.3d %+3.3d\n", dx, dy) ;
	byteNum = 4 ;

      } else if ((byteNum == 4) &&  ((*ptr == 0x20) || (*ptr == 0x00))) {
	//printf("  Byte 4 = 0x%2.2x\n",*ptr) ;
	if (middle != (bit = (*ptr & 0x20))) {
	  middle = bit ;
	  button_state &= ~(0x02) ;
	  button_state |= (*ptr & 0x20)>>4;
	  button_shm->write(&button_state);
	  //printf("middle = %d\n\n\n",middle) ;
	}
	byteNum = 1 ;

      } else {
	printf("\n\nHuh? = 0x%2.2x\n\n\n",*ptr) ;
      }
    }
  } //for(unsigned char *ptr=read_buffer ;ptr < end_ptr ;ptr++)
  
  /* for VRCO (tm) trackd (tm) compatiablity */
#ifdef TRACKD_COMPATABLE
  {
    int i ;
    for(i=0 ;i<TD_BUTTON_ARRAY_NUM ;i++)
      VRCO.button[i] = (button_state & (01 << i)) >> i ;
    VRCO.val[0] = joystick_state[0] ;
    VRCO.val[1] = joystick_state[0] ;
  }
#endif
  return DTKSERVICE_CONTINUE ;
}


// The DTK C++ loader/unloader functions

static dtkService *dtkDSO_loader(const char *arg) {
  printf("%s: %d\n",__FILE__,__LINE__) ;
  return new Vwand ;
}

static int dtkDSO_unloader(dtkService *vwand) {
  delete vwand ;
  return DTKDSO_UNLOAD_CONTINUE ; // success
}
